<?php

namespace Drupal\blazy\Media;

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\image\Plugin\Field\FieldType\ImageItem;
use Drupal\media\MediaInterface;
use Drupal\blazy\Blazy;
use Drupal\blazy\BlazyDefault;
use Drupal\blazy\Utility\Path;
use Drupal\blazy\Utility\Sanitize;

/**
 * Provides image-related methods.
 *
 * @todo recap similiraties and make them plugins.
 */
class BlazyImage {

  /**
   * Checks if the image style contains crop in the effect name.
   *
   * @var array
   */
  private static $crop;

  /**
   * Checks if image dimensions are set.
   *
   * @var array
   */
  private static $isCropSet;

  /**
   * Prepares CSS background image.
   *
   * @todo refactor this, to get rid of settings for blazies object at/ by 3.x.
   */
  public static function background(array $settings, $style = NULL) {
    // @tbd replace src with URL before 3.x, or keep it.
    return [
      'src' => self::url($settings, $style),
      'ratio' => self::ratio($settings),
    ];
  }

  /**
   * Returns the image style if it contains crop effect.
   *
   * @param object $style
   *   The image style to check for.
   *
   * @return object
   *   Returns the image style instance if it contains crop effect, else NULL.
   */
  public static function getCrop($style): ?object {
    $id = $style->id();

    if (!isset(self::$crop[$id])) {
      $output = NULL;

      foreach ($style->getEffects() as $effect) {
        if (strpos($effect->getPluginId(), 'crop') !== FALSE) {
          $output = $style;
          break;
        }
      }
      self::$crop[$id] = $output;
    }
    return self::$crop[$id];
  }

  /**
   * Sets dimensions once to reduce method calls, if image style contains crop.
   *
   * @param array $settings
   *   The settings being modified.
   * @param object $style
   *   The image style to check for crp effect.
   */
  public static function cropDimensions(array &$settings, $style): void {
    $id = $style->id();

    if ($style && !isset(self::$isCropSet[$id])) {
      // If image style contains crop, sets dimension once, and let all inherit.
      if ($crop = self::getCrop($style)) {
        $blazies = $settings['blazies'];
        $data = self::transformDimensions($crop, $blazies);

        // Informs individual images that dimensions are already set once.
        // Do not let the first broken image screw up the rest, likely
        // non-trasliterated file names, SVG, missing ones, etc.
        if ($data['width']) {
          $blazies->set('image', $data, TRUE)
            ->set('is.dimensions', TRUE);
        }
      }

      self::$isCropSet[$id] = TRUE;
    }
  }

  /**
   * Provides original unstyled image dimensions based on the given image item.
   *
   * This one is original image, not styled like self:transformDimensions().
   * Sources: formatters, filters or any hard-coded unmanaged files like VEF.
   */
  public static function dimensions(array &$settings, $item = NULL, $initial = FALSE): array {
    $blazies = $settings['blazies'];
    $_width  = 'width';
    $_height = 'height';
    $fluid   = $blazies->is('fluid');
    $ratios  = $blazies->get('css.ratio');
    $which   = $initial ? 'first' : 'image';
    $height  = $blazies->get($which . '.height');
    $width   = $blazies->get($which . '.width');
    $uri     = $blazies->get($which . '.uri');

    if ($item) {
      $width = $item->width ?? $width;
      $height = $item->height ?? $height;
      $blazies->set('image.item', $item);
    }

    // Only applies when Image style is empty, no file API, no $item,
    // with unmanaged VEF/ WYSIWG/ filter image without image_style.
    if ($uri && empty($settings['image_style']) && (!$height || !$width)) {
      $abs = $blazies->get('image.uri_root', $uri);
      // Must be valid URI, or web-accessible url, not: /modules|themes/...
      if (!BlazyFile::isValidUri($abs) && mb_substr($abs, 0, 1) == '/') {
        if ($request = Path::requestStack()) {
          $abs = $request->getCurrentRequest()->getSchemeAndHttpHost() . $abs;
        }
      }

      // Prevents 404 warning when video thumbnail missing for a reason.
      if ($data = @getimagesize($abs)) {
        [$width, $height] = $data;
      }
    }

    // Sometimes they are string, cast them integer to reduce JS logic.
    // These were the last two standing settings along with URI, now gone
    // for good into blazies object.
    $check[$_width] = $width;
    $check[$_height] = $height;

    self::toInt($check, $_width, $_height);

    // Defines original dimensions.
    $data = ['width' => $check[$_width], 'height' => $check[$_height]];

    if ($fluid) {
      $dim = $data;
      $dim['ratios'] = $ratios;
      $data['fluid'] = self::fluid($dim);
    }

    $data['ratio'] = self::ratio($data);

    // If initial call, used by EZ, etc.
    if ($initial || !$blazies->get('first.width')) {
      $blazies->set('first', $data, TRUE);
    }

    // Only if not cropped uniformly.
    if (!$blazies->is('dimensions')) {
      $blazies->set('image', $data, TRUE);
    }

    // In case `image_style` is not provided.
    $blazies->set('image.original', $data, TRUE);

    return $data;
  }

  /**
   * Returns the image item out of File entity, ER, etc., or just $settings.
   *
   * @param object $object
   *   The optional Media, File entity, or ER, etc. to get image item from.
   * @param array $settings
   *   The optional settings.
   *
   * @return object|null
   *   The object of image item, or NULL.
   *
   * @todo simplify this, like everything else. An obvious confusion here.
   * @todo return image item directly without settings.
   */
  public static function fromAny($object = NULL, array &$settings = []): ?object {
    // @todo remove check at 3.x after sub-modules and VEF removed.
    Blazy::verify($settings);
    $blazies = $settings['blazies'];
    $output = $uri = NULL;

    // If Media entity, we must have a File entity, and likely ImageItem.
    if ($object instanceof MediaInterface) {
      $entity = $object;
    }
    else {
      // Extracts File entity from any object or settings, if applicable.
      // Node, EntityReferenceRevisionsItem, etc.
      $entity = BlazyFile::item($object, $settings);

      // Called by BlazyFilter file upload and legacy BlazyViewsFieldFile.
      if (BlazyFile::isFile($entity)
        && $factory = Blazy::service('image.factory')) {
        // Might be a video/ audio file URI, not just image.
        $check = $entity->getFileUri();

        if ($image = $factory->get($check)) {
          $output = self::fakeFromFactory($blazies, $entity, $image);
          $uri = $output ? ($output->uri ?? NULL) : NULL;
        }
      }
    }

    // Called by formatters.
    if (empty($output)) {
      $options = [
        'entity' => $entity,
        'source' => $entity == $object ? NULL : $object,
        'settings' => $settings,
      ];

      // We may have a Media entity, etc.
      $output = self::fromContent($options);
    }

    // Final URI check.
    $uri = $uri ?: BlazyFile::uri($output, $settings);

    if ($uri) {
      $blazies->set('image.uri', $uri);

      // Prepare image URL and its dimensions, including for rich-media content,
      // such as for local video poster image if a poster URI is provided.
      // Url needs to be defined here for file_video.
      // @todo move it back to #pre_render, once file_video poster is resolved.
      self::prepare($settings, $output, $uri);
    }

    return $output;
  }

  /**
   * Returns TRUE if an ImageItem.
   */
  public static function isImage($item): bool {
    return $item instanceof ImageItem;
  }

  /**
   * Returns the image item from any sources, if available.
   *
   * PHP 7.2 accepts object. D8 >= PHP 7.3. Not good for D7 backport.
   */
  public static function item($item = NULL, array $options = [], $name = NULL): ?object {
    return self::isImage($item) ? $item : self::fromContent($options, $name);
  }

  /**
   * Returns the image item from any sources, if available.
   *
   * This block is a bit scary yet it is a more organized way to extract Image
   * item from various sources in tandem with custom settings.image previously
   * scattered with if-else. This has saved more than 60 lines, and two methods:
   * ::fromMedia(), already gone. Can be better.
   */
  public static function fromContent(array $options, $name = NULL): ?object {
    $settings = Blazy::toHashtag($options);
    $blazies  = $settings['blazies'] ?? NULL;
    $poster   = $settings['image'] ?? NULL;
    $name     = $name ?: $poster;

    // If poster is not defined, use the source_field or thumbnail property.
    // Title is NULL from thumbnail, likely core bug, so use source.
    if ($blazies && !$name && $source = $blazies->get('media.source')) {
      $name = $source == 'image' ? $blazies->get('media.source_field') : 'thumbnail';
    }

    $func = function ($key, $property) use ($options) {
      $object = ($options[$key] ?? NULL);
      if ($object instanceof ContentEntityInterface
        && $object->hasField($property)) {
        $item = $object->get($property)->first();
        $valid = self::isImage($item);

        // Media embedded inside Paragraph item as defined by settings.image,
        // basically drilling down nested entities here to find the gold.
        if ($item) {
          if (!$valid && $entity = ($item->entity ?? NULL)) {
            if ($entity instanceof ContentEntityInterface
              && $entity->hasField('thumbnail')) {
              $item = $entity->get('thumbnail')->first();
              $valid = self::isImage($item);
            }
          }

          // For Remote video, it has meaningful label from OEmbed, OOTB.
          /* @phpstan-ignore-next-line */
          if ($valid && trim($item->title ?? '') == '') {
            /* @phpstan-ignore-next-line */
            $item->title = $object->label();
          }
        }

        return $valid ? $item : NULL;
      }
      return NULL;
    };

    // \Drupal\paragraphs\Entity\Paragraph, Media, Node, etc.
    $item = $func('entity', $name) ?: $func('source', $name);
    $item = $name ? $item : NULL;
    if (!$item) {
      $item = $func('entity', 'thumbnail') ?: $func('source', 'thumbnail');
    }

    return $item;
  }

  /**
   * Disable image style if so configured.
   *
   * Extensions without image styles: animated GIF, APNG, SVG, etc.
   *
   * @requires CheckItem::essentials()
   */
  public static function checkUnstyled(array &$settings, $uri, $first_time = FALSE): bool {
    $blazies = $settings['blazies'];
    $ext = pathinfo($uri, PATHINFO_EXTENSION);
    $external = UrlHelper::isExternal($uri);
    $extensions = ['svg'];

    // If we have added extensions.
    if ($unstyles = $blazies->ui('unstyled_extensions')) {
      $checks = array_map('trim', explode(' ', strtolower($unstyles)));
      $checks = array_merge($checks, $extensions);
      $extensions = array_unique($checks);
    }

    $unstyled = $ext && in_array($ext, $extensions);
    if (!$unstyled) {
      // @todo recheck if anything against this at all.
      $unstyled = $external || Blazy::isDataUri($uri);
    }

    // Re-define, if the provided API by-passed, or different/ altered per item.
    if ($first_time) {
      $blazies->set('is.external', $external)
        ->set('is.unstyled', $unstyled)
        ->set('image.extension', $ext);
    }

    return $unstyled;
  }

  /**
   * Checks if we have image item.
   *
   * Both ImageItem and fake stdClass are valid, no problem.
   */
  public static function isValidItem($item): bool {
    $item = is_array($item) ? Blazy::toHashtag($item, 'item', NULL) : $item;
    return is_object($item) && (isset($item->uri) || isset($item->target_id));
  }

  /**
   * Prepares URLs, placeholder, and dimensions for an individual image.
   *
   * Respects a few scenarios:
   * 1. Blazy Filter or unmanaged file with/ without valid URI.
   * 2. Hand-coded image_url with/ without valid URI.
   * 3. Respects first_uri without image_url such as colorbox/zoom-like.
   * 4. File API via field formatters or Views fields/ styles with valid URI.
   * If we have a valid URI, provides the correct image URL.
   * Otherwise leave it as is, likely hotlinking to external/ sister sites.
   * Hence URI validity is not crucial in regards to anything but #4.
   * The image will fail silently at any rate given non-expected URI.
   *
   * @param array $settings
   *   The given settings being modified.
   * @param object $item
   *   The image item.
   * @param string $uri
   *   The image uri.
   *
   * @requires self::checkUnstyled()
   */
  public static function prepare(array &$settings, $item = NULL, $uri = NULL): void {
    // @fixme figure out why another reset is required here to get correct item,
    // otherwise the below bailout called once for entire items. No big deal to
    // renew, just not as expected. Strangely only happens sometimes at
    // some modules override, some edge cases.
    // @todo remove this reset once things work consistently at all cases.
    $blazies = $settings['blazies']->reset($settings);
    $uri = $blazies->get('image.uri') ?: $uri;

    // Bail out if already processed.
    // @fixme called once for the entire items, not just this context, unless
    // reset as above, see dup zooming samples. Only some edge cases, though.
    if ($blazies->was('url') && $blazies->get('image.url')) {
      return;
    }

    // Might be called from Views without Blazy formatter, like Image formatter.
    // Since Blazy:2.9, image style entity is loaded once at container level,
    // but might still be needed for adopted Image formatter by a Views style.
    // @todo since done at container, it might also truble the unstyled per URI.
    $style = $blazies->get('image.style');

    // BlazyFilter, or image style with crop, may already set these.
    self::dimensions($settings, $item, FALSE);

    // Define styles regardless unstyled so to have correct dimensions at
    // lightboxes, thumbnails, etc.
    if ($uri) {
      self::styleData($settings, $uri);
    }

    // SVG, APNG, etc. should not use image_style as they don't convert.
    // @todo remove self::unstyled($settings);
    // Provides image url based on the given settings.
    if ($style && !$blazies->is('unstyled')) {
      $blazies->set('cache.metadata.tags', $style->getCacheTags(), TRUE);

      // Only re-calculate dimensions if not cropped, nor already set.
      $resimage = $settings['responsive_image_style'] ?? NULL;
      if (!$blazies->is('dimensions') && !$resimage) {
        $data = self::transformDimensions($style, $blazies, $uri);

        $blazies->set('image', $data, TRUE);
      }
    }

    // Currently doesn't affect option.ratio, a failsafe for BG, else collapsed.
    $url = self::url($settings, $style, $uri);
    $blazies->set('image.url', $url)
      ->set('was.url', TRUE);
  }

  /**
   * Checks defined image styles at one place.
   *
   * @requires \Drupal\blazy\Media\BlazyImage::styles()
   *
   * Image styles were provided once at the container level, but not dimensions
   * which may require URIs at item level. Previously these are scattered around
   * as required, now called once for all. Nothing loaded if not so configured.
   *
   * @todo remove `image` check after another check. Was needed to be undefined
   * to not conflict with Responsive image last time, till required. Also image
   * may be set once if cropped at self::cropDimensions().
   * URI is not available at container level, except for the first,
   * or when preload option is enabled, unless enforced in the far future.
   */
  public static function styleData(array &$settings, $uri): void {
    $blazies  = $settings['blazies'];
    $unstyled = $blazies->is('unstyled');

    foreach (BlazyDefault::imageStyles() as $key) {
      // @todo re-enable the skip if any issues with Responsive image.
      // The skip limits SVG dimension checks, ratio, url, etc.
      // if ($key == 'image') {
      // continue;
      // }
      if ($style = $blazies->get($key . '.style')) {
        $data = self::transformDimensions($style, $blazies, $uri);
        $blazies->set($key, $data, TRUE);

        $url = self::url($settings, $style, $uri);
        $blazies->set($key . '.url', $url);
      }
    }

    // ResponsiveImage is the most temperamental module. Unlike plain old Image,
    // it explodes when the image is missing as much as when fed wrong URI, etc.
    // Do not let SVG alike mess up with ResponsiveImage, else fatal.
    if ($style = BlazyResponsiveImage::toStyle($settings, $unstyled)) {
      $blazies->set('resimage.style', $style);

      // Might be set via BlazyFilter, but not enough data passed.
      $multiple = $blazies->is('multistyle');
      if (!$blazies->get('resimage.id') || $multiple) {
        BlazyResponsiveImage::define($blazies, $style);
      }

      // We'll bail out internally if already set once at container level.
      BlazyResponsiveImage::dimensions($settings, $style, FALSE);
    }

    // Checks SVG dimensions, if any.
    if (BlazyFile::isValidUri($uri) && $blazies->get('image.extension') == 'svg') {
      if ($blazies->use('svg_dimensions')) {
        $svg = simplexml_load_file($uri);
        if ($svg && isset($svg['viewBox'])) {
          [,, $width, $height] = array_map('trim', explode(' ', $svg['viewBox']));

          if ($width && $height) {
            $blazies->set('image.width', ceil($width))
              ->set('image.height', ceil($height));
          }
        }
      }
      else {
        $blazies->set('image.width', NULL)
          ->set('image.height', NULL);
      }
    }
  }

  /**
   * Provides a computed image ratio aka fluid ratio.
   */
  public static function fluid(array $data): ?string {
    $width  = $data['width'];
    $height = $data['height'];
    $ratios = $data['ratios'] ?? BlazyDefault::RATIO;
    $output = NULL;

    if (empty($width) || empty($height)) {
      return $output;
    }

    $width  = (int) $width;
    $height = (int) $height;

    try {
      $check  = self::toRatio($width, $height);
      $result = ($width / $check) . ':' . ($height / $check);

      if (in_array($result, $ratios)) {
        $output = $result;
      }
    }
    catch (\DivisionByZeroError $e) {
      // Do nothing, optional features should not mess up the rest.
    }
    catch (\Exception $e) {
      // Do nothing also.
    }

    return $output;
  }

  /**
   * Provides a computed image ratio aka fluid ratio.
   *
   * Addresses multi-image-style Responsive image or, plain old one.
   * A failsafe for BG, else collapsed.
   *
   * @todo decide if to provide NULL or 0 instead.
   * @todo converts to blazies at/by 3.x.
   */
  public static function ratio(array $data) {
    $no_dims = empty($data['height']) || empty($data['width']);
    return $no_dims ? 100 : round((($data['height'] / $data['width']) * 100), 2);
  }

  /**
   * Checks for Image styles at container level once, except for multi-styles.
   *
   * Specific for lightbox, it can also be Responsive image, but not here.
   *
   * @param array $settings
   *   The modified settings.
   * @param bool $multiple
   *   A flag for various Image styles: Blazy Filter, etc., old GridStack.
   *   While most field formatters can only have one image style per field.
   */
  public static function styles(array &$settings, $multiple = FALSE): void {
    $blazies = $settings['blazies'];
    if ($manager = Blazy::service('blazy.manager')) {
      foreach (BlazyDefault::imageStyles() as $key) {
        if (!$blazies->get($key . '.style') || $multiple) {
          if ($_style = ($settings[$key . '_style'] ?? '')) {
            if ($entity = $manager->load($_style, 'image_style')) {
              $blazies->set($key . '.style', $entity)
                ->set($key . '.id', $entity->id());
            }
          }
        }
      }
    }
  }

  /**
   * Returns the thumbnail image using theme_image(), or theme_image_style().
   *
   * Alt and SRC will be auto-escaped when entering Twig, this is just to make
   * sure no unknown edge cases get in the way.
   *
   * @see https://www.drupal.org/node/2489544
   */
  public static function thumbnail(array $settings, $item = NULL): array {
    $blazies = $settings['blazies'];
    $uri     = $blazies->get('thumbnail.uri') ?: $blazies->get('image.uri');

    if (!$uri) {
      return [];
    }

    $unstyled = $blazies->is('unstyled');
    $style    = $blazies->get('thumbnail.id') ?: $settings['thumbnail_style'] ?? NULL;
    $alt      = $blazies->get('image.alt');
    $valid    = $blazies->get('image.valid') ?: BlazyFile::isValidUri($uri);

    // Thumbnails can use image styles, except for SVG for now.
    // @todo check for any modules (ImageMagick) which convert SVG to image,
    // and remove this check if present, leaving it for external URL + data URI.
    if ($valid && $blazies->get('image.extension') != 'svg') {
      $unstyled = FALSE;
    }

    return [
      '#theme'      => $unstyled ? 'image' : 'image_style',
      '#style_name' => $style ?: 'thumbnail',
      '#uri'        => $valid ? $uri : UrlHelper::stripDangerousProtocols($uri),
      '#item'       => $item,
      '#alt'        => $alt ? Html::escape(strip_tags($alt)) : '',
    ];
  }

  /**
   * A wrapper for ImageStyle::transformDimensions().
   *
   * @param object $style
   *   The given image style.
   * @param array|object $config
   *   The data config: width, height, and uri, or $blazies as config source.
   * @param string $uri
   *   The optional URI if differs from main image, such as thumbnail URI.
   */
  public static function transformDimensions($style, $config, $uri = NULL): array {
    $fluid  = FALSE;
    $ratios = [];

    // Default non-API source:
    if (is_array($config)) {
      $uri    = $uri ?: ($config['uri'] ?? '');
      $width  = $config['width'] ?? NULL;
      $height = $config['height'] ?? NULL;
    }
    // A convenient API source, must be original sizes:
    else {
      $fluid  = $config->is('fluid');
      $ratios = $config->get('css.ratio');
      $uri    = $uri ?: ($config->get('image.uri') ?: $config->get('first.uri'));
      $width  = $config->get('image.original.width') ?: $config->get('first.width');
      $height = $config->get('image.original.height') ?: $config->get('first.height');
    }

    $dim = ['width' => $width, 'height' => $height];

    // Funnily $uri is ignored at all core image effects.
    $style->transformDimensions($dim, $uri);

    // Sometimes they are string, cast them integer to reduce JS logic.
    self::toInt($dim, 'width', 'height');

    if ($fluid) {
      $info = $dim;
      $info['ratios'] = $ratios;
      $fluid = self::fluid($info);
    }

    // Keys here are hard-coded, so to be inherited by children as intended.
    // See self::dimensions().
    return [
      'width'  => $dim['width'],
      'height' => $dim['height'],
      'ratio'  => self::ratio($dim),
      'fluid'  => $fluid,
    ];
  }

  /**
   * Returns image URL with an optional image style.
   *
   * Addressed various sources:
   * - URL which should not be styled: animated gif, apng, svg, etc.
   * - UGC image URL, with likely invalid URI due to hard-coded markdown, etc.
   * - Responsive image vs. regular image style.
   *
   * @requires \Drupal\blazy\BlazyInternal::prepare()
   *
   * @see self::prepare()
   * @see self::background()
   * @see BlazyResponsiveImage::background()
   *
   * @todo remove fallbacks after another check, also settings after migration.
   */
  public static function url(array $settings, $style = NULL, $uri = NULL) {
    $blazies = $settings['blazies'];
    $uri     = $uri ?: $blazies->get('image.uri');
    $valid   = BlazyFile::isValidUri($uri);
    $styled  = $valid && !$blazies->is('unstyled');
    $style   = $styled ? $style : NULL;
    $url     = $settings['image_url'] ?? '';
    $url     = $blazies->get('image.url') ?: $url;
    $options = ['url' => $url];

    $url = BlazyFile::transformRelative($uri, $style, $options);

    // Just in case, an attempted kidding gets in the way, relevant for UGC.
    // @todo re-check to completely remove data URI.
    if ($url && $blazies->is('unsafe')) {
      $use_data_uri = $blazies->filter('use_data_uri');
      $url = Sanitize::url($url, $use_data_uri);
    }

    return $url;
  }

  /**
   * Returns fake image item based on the given $blazies.
   */
  public static function fakeFromSettings($blazies) {
    $item = new \stdClass();
    foreach (BlazyDefault::imageProperties() as $key) {
      if ($value = $blazies->get('image.' . $key)) {
        $item->{$key} = $value;
      }
    }
    return $item;
  }

  /**
   * Returns data to provide fake image item of file entity via ImageFactory.
   */
  private static function fakeFromFactory(&$blazies, $file, $image): ?object {
    /** @var \Drupal\file\Entity\File $file */
    [$type] = explode('/', $file->getMimeType(), 2);

    // Including image/svg+xml.
    if ($type == 'image' && $image->isValid()) {
      $data = [
        'uri'       => $file->getFileUri(),
        'target_id' => $file->id(),
        'width'     => $image->getWidth(),
        'height'    => $image->getHeight(),
        'alt'       => $file->getFilename(),
        'title'     => $file->getFilename(),
        'type'      => 'image',
        'entity'    => $file,
      ];

      $blazies->set('image', $data, TRUE);
      return self::fakeFromSettings($blazies);
    }
    return NULL;
  }

  /**
   * Converts dimensions to integer unless empty.
   */
  private static function toInt(array &$data, $width, $height): void {
    $data[$width] = empty($data[$width]) ? NULL : (int) $data[$width];
    $data[$height] = empty($data[$height]) ? NULL : (int) $data[$height];
  }

  /**
   * Provides a computed image ratio aka fluid ratio.
   */
  private static function toRatio($width, $height) {
    if ($width == 0 || $height == 0) {
      return abs(max(abs($width), abs($height)));
    }

    $result = $width % $height;
    return ($result != 0) ? self::toRatio($height, $result) : abs($height);
  }

}
